# Copyright 2021 Le Filament (<http://www.le-filament.com>)
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).

from dateutil.relativedelta import relativedelta

from odoo import api, fields, models
from odoo.tools import date_utils


class AccOperation(models.Model):
    _inherit = "acc.operation"

    # ------------------------------------------------------
    # Fields declaration
    # ------------------------------------------------------
    is_account_active = fields.Boolean(
        "Facturation activée", compute="_compute_is_account_active", store=True
    )
    is_surplus_active = fields.Boolean(
        "Facturation Surplus activée", compute="_compute_is_surplus_active", store=True
    )
    acc_sale_price_ids = fields.One2many("acc.sale.price", "acc_operation_id")

    acc_sale_price_conf_ids = fields.One2many(
        "acc.price.conf",
        "acc_operation_id",
        domain=[("type", "=", "sale")],
        string="Liste des Prix de Vente",
    )
    acc_buyer_price_conf_ids = fields.One2many(
        "acc.price.conf",
        "acc_operation_id",
        domain=[("type", "=", "buyer")],
        string="Liste des Prix d'Achat",
    )
    acc_accise_price_conf_ids = fields.One2many(
        "acc.price.conf",
        "acc_operation_id",
        domain=[("type", "=", "accise")],
        string="Liste des Tarifs ACCISE",
    )

    acc_sale_price_surplus_ids = fields.One2many(
        "acc.sale.price.buyer", "acc_operation_id"
    )
    acc_account_ids = fields.One2many(
        "acc.account",
        "acc_operation_id",
        domain=[("is_account_buyer", "=", False)],
        string="Factures",
    )
    acc_account_surplus_ids = fields.One2many(
        "acc.account",
        "acc_operation_id",
        domain=[("is_account_buyer", "=", True)],
        string="Factures Surplus",
    )
    acc_account_line_ids = fields.One2many("acc.account.line", "acc_operation_id")
    acc_account_count = fields.Integer(
        string="Nombre de factures", compute="_compute_acc_account_count"
    )
    acc_account_surplus_count = fields.Integer(
        string="Nombre de factures Surplus",
        compute="_compute_acc_account_surplus_count",
    )
    sale_price = fields.Float("Prix de vente par défaut", digits="Sale Price")
    price_accise = fields.Float("Tarif ACCISE par défaut", digits="Sale Price")
    # ------------------------------------------------------
    # SQL Constraints
    # ------------------------------------------------------

    # ------------------------------------------------------
    # Default methods
    # ------------------------------------------------------

    # ------------------------------------------------------
    # Computed fields / Search Fields
    # ------------------------------------------------------
    @api.depends("acc_account_ids")
    def _compute_acc_account_count(self):
        """
        Fonction de calculer le nombre de factures de vente liées à une opération
        """
        for res in self:
            res.acc_account_count = len(res.acc_account_ids)

    @api.depends("acc_account_surplus_ids")
    def _compute_acc_account_surplus_count(self):
        """
        Fonction de calculer le nombre de factures d'Achat liées à une opération
        """
        for res in self:
            res.acc_account_surplus_count = len(res.acc_account_surplus_ids)

    @api.depends("acc_injection_ids", "acc_injection_ids.is_account_auto_activate")
    def _compute_is_account_active(self):
        """
        Fonction permettant de vérifier si une opération possède la facturation
        de vente activée. L'info est récupérée de chaque point d'injection.
        Si un point d'injection a la facturation de vente activée,
        alors l'opération aussi.
        """
        for res in self:
            count_ids = res.acc_injection_ids.filtered(
                lambda r: r.is_account_auto_activate is True
            )
            if count_ids:
                res.is_account_active = True
            else:
                res.is_account_active = False

    @api.depends("acc_injection_ids", "acc_injection_ids.is_account_surplus_activate")
    def _compute_is_surplus_active(self):
        """
        Fonction permettant de vérifier si une opération possède la facturation
        d'achat activée. L'info est récupérée de chaque point d'injection.
        Si un point d'injection a la facturation d'achat activée,
        alors l'opération aussi.
        """
        for res in self:
            count_ids = res.acc_injection_ids.filtered(
                lambda r: r.is_account_surplus_activate is True
            )
            if count_ids:
                res.is_surplus_active = True
            else:
                res.is_surplus_active = False

    # ------------------------------------------------------
    # Onchange / Constraints
    # ------------------------------------------------------

    # ------------------------------------------------------
    # CRUD methods (ORM overrides)
    # ------------------------------------------------------

    # ------------------------------------------------------
    # Actions
    # ------------------------------------------------------
    def action_view_acc_accounts(self):
        """
        Action pour accéder aux factures de vente liées à l'opération
        """
        self.ensure_one()
        action = self.env["ir.actions.actions"]._for_xml_id(
            "acc_account.action_acc_account"
        )
        action["domain"] = [
            ("acc_operation_id", "=", self.id),
            ("is_account_buyer", "=", False),
        ]
        return action

    def action_view_acc_accounts_surplus(self):
        """
        Action pour accéder aux factures d'achat liées à l'opération
        """
        self.ensure_one()
        action = self.env["ir.actions.actions"]._for_xml_id(
            "acc_account.action_acc_account"
        )
        action["domain"] = [
            ("acc_operation_id", "=", self.id),
            ("is_account_buyer", "=", True),
        ]
        return action

    def update_price_sale(self):
        """Mise à jour du prix de vente en masse sur les points de soutirage"""
        for op in self:
            if op.sale_price:
                op.acc_delivery_ids.write({"sale_price": op.sale_price})

    def update_price_accise(self):
        """Mise à jour des tarifs ACCISE en masse sur les points de soutirage"""
        for op in self:
            if op.price_accise:
                op.acc_delivery_ids.write({"price_accise": op.price_accise})

    def get_power_by_cons(self, acc_delivery_id, date_start, date_end):
        """
        Fonction retournant l'énergie consommée d'un consommateur donné pour une
        période donnée.
        :param  acc_delivery_id: consommateur
                date_start: date début
                date_end: date de fin
        :return: un dictionnaire de valeurs (consommateur + power + type)
        """
        power_tab_by_cons = self.env["acc.enedis.cdc"].read_group(
            [
                ("acc_operation_id", "=", self.id),
                ("acc_counter_id", "=", acc_delivery_id.id),
                ("date_slot", ">=", date_start),
                ("date_slot", "<=", date_end),
                ("comp_data_type", "=", "autocons"),
            ],
            ["power", "acc_operation_id", "date_slot"],
            ["acc_counter_id"],
            orderby="acc_counter_id",
            lazy=False,
        )

        return power_tab_by_cons

    def get_power_by_prod(self, acc_injection_id, date_start, date_end):
        """
        Fonction retournant l'énergie produite et en surplus d'un producteur donné
        pour une période donnée.
        :param  acc_injection_id: producteur
                date_start: date début
                date_end: date de fin
        :return: un dictionnaire de valeurs (producteur + power + type)
        """
        power_tab_by_prod = self.env["acc.enedis.cdc"].read_group(
            [
                ("acc_operation_id", "=", self.id),
                ("acc_counter_id", "=", acc_injection_id.id),
                ("date_slot", ">=", date_start),
                ("date_slot", "<=", date_end),
                "|",
                ("comp_data_type", "=", "prod"),
                ("comp_data_type", "=", "surplus"),
            ],
            ["power", "acc_operation_id", "date_slot"],
            ["comp_data_type"],
            orderby="comp_data_type",
            lazy=False,
        )

        return power_tab_by_prod

    def process_create_account(
        self, acc_injection_id, acc_delivery_id, date_start, date_end, power_total
    ):
        """
        Fonction permettant la création des factures de vente,
        ainsi que ces lignes de facture entre un producteur et un consommateur
        pour une période donnée.
        :param  acc_injection_id: un producteur
                acc_delivery_id: un consommateur
                date_start: date début
                date_end: date de fin
                power_total: puissance totale sur la période
        """
        account_list = []
        Account = self.env["acc.account"]

        acc_account = Account.create(
            {
                "acc_operation_id": self.id,
                "acc_injection_id": acc_injection_id.id,
                "acc_delivery_id": acc_delivery_id.id,
                "power_cons": power_total,
                "start_date": date_start,
                "end_date": date_end,
            }
        )

        interval_ids = self.env["acc.price.conf"].search(
            [
                ("acc_operation_id", "=", self.id),
                ("acc_injection_id", "=", acc_injection_id.id),
                ("acc_delivery_id", "=", acc_delivery_id.id),
                ("start_date", "<=", date_start),
                ("type", "=", "sale"),
                ("end_date", "=", False),
            ]
        )
        if not interval_ids:
            price_ids = self.env["acc.price.conf"].search(
                [
                    ("type", "=", "sale"),
                    ("acc_operation_id", "=", self.id),
                    ("acc_injection_id", "=", acc_injection_id.id),
                    ("acc_delivery_id", "=", acc_delivery_id.id),
                    "|",
                    ("end_date", ">=", date_start),
                    ("end_date", "=", False),
                ]
            )
            price2_ids = self.env["acc.price.conf"].search(
                [
                    ("type", "=", "sale"),
                    ("acc_operation_id", "=", self.id),
                    ("acc_injection_id", "=", acc_injection_id.id),
                    ("acc_delivery_id", "=", acc_delivery_id.id),
                    ("start_date", "<=", date_end),
                ]
            )
            interval_ids = price_ids & price2_ids

        # Création des lignes de facturation
        # pour chaque période de tarifs trouvée
        for interval in interval_ids:
            if date_start > interval.start_date:
                date_interval_start = date_start
            else:
                date_interval_start = interval.start_date

            if interval.end_date:
                if date_end < interval.end_date:
                    date_interval_end = date_end
                else:
                    date_interval_end = interval.end_date
            else:
                date_interval_end = date_end

            power_total = self.calc_power_by_cons_by_prod(
                acc_injection_id,
                acc_delivery_id,
                date_interval_start,
                date_interval_end,
            )

            AccountLine = self.env["acc.account.line"]

            # Création de la ligne de facture
            # par intervalle de période de prix
            AccountLine.create(
                {
                    "quantity": power_total,
                    "price_unit": interval.price,
                    "acc_account_id": acc_account.id,
                    "start_date": date_interval_start,
                    "end_date": date_interval_end,
                    "description": "Période du "
                    + str(date_interval_start)
                    + " au "
                    + str(date_interval_end),
                }
            )
            account_list.append(acc_account)

        tax_interval_ids = self.env["acc.account.tax.tcfe"].search(
            [
                ("start_date", "<=", date_start),
                ("end_date", "=", False),
            ]
        )
        if not tax_interval_ids:
            tax_ids = self.env["acc.account.tax.tcfe"].search(
                [
                    ("end_date", ">=", date_start),
                ]
            )
            tax2_ids = self.env["acc.account.tax.tcfe"].search(
                [
                    ("start_date", "<=", date_end),
                ]
            )
            tax_interval_ids = tax_ids & tax2_ids

        # Création des lignes de facturation
        # pour chaque période de taxes trouvée
        for interval in tax_interval_ids:
            if date_start > interval.start_date:
                date_interval_start = date_start
            else:
                date_interval_start = interval.start_date

            if interval.end_date:
                if date_end < interval.end_date:
                    date_interval_end = date_end
                else:
                    date_interval_end = interval.end_date
            else:
                date_interval_end = date_end

            power_total = self.calc_power_by_cons_by_prod(
                acc_injection_id,
                acc_delivery_id,
                date_interval_start,
                date_interval_end,
            )

            AccountLine = self.env["acc.account.line"]

            # Création de la ligne de facture
            # par intervalle de période de prix
            AccountLine.create(
                {
                    "quantity": power_total,
                    "is_tax": True,
                    "price_unit": interval.price,
                    "acc_account_id": acc_account.id,
                    "start_date": date_interval_start,
                    "end_date": date_interval_end,
                    "description": "Période du "
                    + str(date_interval_start)
                    + " au "
                    + str(date_interval_end),
                }
            )
            account_list.append(acc_account)
        return True

    def create_account(self, date_month, account_periodicity):
        """
        Fonction permettant de générer les factures à une date donné.
        Création des factures + Création des lignes de facture associées.
        :param  date_month: date à laquelle on lance la facture
                account_periodicity: périodicité (1 pour mois, 3 pour trimestre,
                6 pour semestre, 12 pour année)
        :return:
        """
        date_end = date_utils.end_of(date_month, "month")
        if account_periodicity == "1":
            date_start = date_utils.start_of(date_month, "month")
        else:
            date_start = date_end - relativedelta(months=int(account_periodicity) - 1)
            date_start = date_utils.start_of(date_start, "month")

        for acc_injection_id in self.acc_injection_ids:
            if acc_injection_id.is_account_auto_activate:
                for acc_delivery_id in self.acc_delivery_ids:

                    power_total = self.calc_power_by_cons_by_prod(
                        acc_injection_id, acc_delivery_id, date_start, date_end
                    )
                    if power_total > 0:
                        # Création du lot
                        batch_name = (
                            "Factures du "
                            + str(date_start)
                            + " au "
                            + str(date_end)
                            + " - Opération :"
                            + str(self.name)
                            + " - Prod :"
                            + str(acc_injection_id.name)
                            + " - Conso :"
                            + str(acc_delivery_id.name)
                        )
                        batch = self.env["queue.job.batch"].get_new_batch(batch_name)

                        self.with_context(
                            job_batch=batch
                        ).with_delay().process_create_account(
                            acc_injection_id,
                            acc_delivery_id,
                            date_start,
                            date_end,
                            power_total,
                        )
                        # lancement de la file d'attente
                        batch.enqueue()

        action = self.env["ir.actions.actions"]._for_xml_id(
            "acc_account.action_acc_account"
        )
        action["domain"] = [("acc_operation_id", "=", self.id)]
        return action

    def calc_power_by_cons_by_prod(
        self, acc_injection_id, acc_delivery_id, date_start, date_end
    ):
        """
        Fonction permettant de calculer la puissance en kWh consommée
        pour un consommateur donnée en fonction
        en rapport à un producteur défini pour un intervalle de temps donné.
        :param  acc_injection_id: producteur
                acc_delivery_id: consommateur
                date_start: date début
                date_end: date de fin
        :return: une valeur représentant la puissance consommée en kWh
        """
        power_tab_by_prod = self.env["acc.enedis.cdc"].read_group(
            [
                ("acc_operation_id", "=", self.id),
                ("date_slot", ">=", date_start),
                ("date_slot", "<=", date_end),
                "|",
                ("comp_data_type", "=", "prod"),
                ("comp_data_type", "=", "surplus"),
            ],
            ["power", "acc_operation_id", "date_slot"],
            ["comp_data_type"],
            orderby="comp_data_type",
            lazy=False,
        )

        if power_tab_by_prod[0]["power"] and power_tab_by_prod[1]["power"]:
            prod_tot = power_tab_by_prod[0]["power"] - power_tab_by_prod[1]["power"]

        power_tab_by_prod = self.get_power_by_prod(
            acc_injection_id, date_start, date_end
        )

        power_total = 0
        if prod_tot > 0:
            power_autoprod_percent = (
                (power_tab_by_prod[0]["power"] - power_tab_by_prod[1]["power"]) * 100
            ) / prod_tot
            power_tab_by_cons = self.get_power_by_cons(
                acc_delivery_id, date_start, date_end
            )
            if power_tab_by_cons[0]["power"]:
                power_total = (power_tab_by_cons[0]["power"] / 1000 / 2) * (
                    power_autoprod_percent / 100
                )
        return power_total

    def process_create_account_surplus(
        self, injection_id, date_start, date_end, power_surplus
    ):
        """
        Fonction permettant la création des factures d'achat,
        ainsi que ces lignes de facture entre un producteur et un acheteur surplus
        pour une période donnée.
        :param  acc_injection_id: un producteur
                acc_delivery_id: un consommateur
                date_start: date début
                date_end: date de fin
                power_total: puissance totale sur la période
        """
        account_list = []
        Account = self.env["acc.account"]

        acc_account = Account.create(
            {
                "acc_operation_id": self.id,
                "buyer_id": injection_id.buyer_id.id,
                "acc_injection_id": injection_id.id,
                "power_cons": power_surplus[1]["power"] / 1000 / 2,
                "start_date": date_start,
                "end_date": date_end,
                # 'price_kwh': price_kwh,
                "is_account_buyer": True,
            }
        )

        interval_ids = self.env["acc.price.conf"].search(
            [
                ("type", "=", "buyer"),
                ("acc_operation_id", "=", self.id),
                ("acc_injection_id", "=", injection_id.id),
                ("acc_buyer_id", "=", injection_id.buyer_id.id),
                ("start_date", "<=", date_start),
                ("end_date", "=", False),
            ]
        )
        if not interval_ids:
            price_ids = self.env["acc.price.conf"].search(
                [
                    ("type", "=", "buyer"),
                    ("acc_operation_id", "=", self.id),
                    ("acc_injection_id", "=", injection_id.id),
                    ("acc_buyer_id", "=", injection_id.buyer_id.id),
                    ("end_date", ">=", date_start),
                ]
            )
            price2_ids = self.env["acc.price.conf"].search(
                [
                    ("type", "=", "buyer"),
                    ("acc_operation_id", "=", self.id),
                    ("acc_injection_id", "=", injection_id.id),
                    ("acc_buyer_id", "=", injection_id.buyer_id.id),
                    ("start_date", "<=", date_end),
                ]
            )
            interval_ids = price_ids & price2_ids

        for interval in interval_ids:
            if date_start > interval.start_date:
                date_interval_start = date_start
            else:
                date_interval_start = interval.start_date

            if interval.end_date:
                if date_end < interval.end_date:
                    date_interval_end = date_end
                else:
                    date_interval_end = interval.end_date
            else:
                date_interval_end = date_end

            power_surplus = self.get_power_by_prod(
                injection_id, date_interval_start, date_interval_end
            )
            AccountLine = self.env["acc.account.line"]

            # Création de la ligne du mois
            AccountLine.create(
                {
                    "quantity": power_surplus[1]["power"] / 1000 / 2,
                    "price_unit": interval.price,
                    "acc_account_id": acc_account.id,
                    "start_date": date_interval_start,
                    "end_date": date_interval_end,
                    "description": "Surplus kWh en ACC du "
                    + str(date_interval_start)
                    + " au "
                    + str(date_interval_end),
                }
            )
            account_list.append(acc_account)
        return True

    def create_account_surplus(self, date_month, account_periodicity):
        """
        Fonction permettant de générer les factures se surplus à une date donnée.
        Création des factures + Création des lignes de facture associées.
        :param  date_month: date à laquelle on lance la facture
                account_periodicity: périodicité (1 pour mois, 3 pour trimestre,
                6 pour semestre, 12 pour année)
        :return:
        """

        date_end = date_utils.end_of(date_month, "month")
        date_start = date_end - relativedelta(months=int(account_periodicity) - 1)
        date_start = date_utils.start_of(date_start, "month")

        for injection_id in self.acc_injection_ids:

            if injection_id.is_account_surplus_activate:
                power_surplus = self.get_power_by_prod(
                    injection_id, date_start, date_end
                )

                # Création du lot
                batch_name = (
                    "Facture surplus du "
                    + str(date_start)
                    + " au "
                    + str(date_end)
                    + " - Opération :"
                    + str(self.name)
                    + " - Prod :"
                    + str(injection_id.name)
                    + " - Acheteur :"
                    + str(injection_id.buyer_id.name)
                )
                batch = self.env["queue.job.batch"].get_new_batch(batch_name)
                self.with_context(
                    job_batch=batch
                ).with_delay().process_create_account_surplus(
                    injection_id, date_start, date_end, power_surplus
                )
                # lancement de la file d'attente
                batch.enqueue()

        action = self.env["ir.actions.actions"]._for_xml_id(
            "acc_account.action_acc_account"
        )
        action["domain"] = [("acc_operation_id", "=", self.id)]
        return action

    def _open_x2m_matrix(self, view_xmlid):
        """
        Fonction qui appelle le wizard matrix 2D pour la défintion des prix
        par consommateur/producteur
        :param  view_xmlid: Vue appelée
        :return: action
        """
        wiz = self.env["acc.price.conf.wizard"].create({})
        view_id = self.env.ref(
            "acc_account.%s" % view_xmlid,
        ).id
        return {
            "name": "Try x2many 2D matrix widget",
            "type": "ir.actions.act_window",
            "view_type": "form",
            "view_mode": "form",
            "res_model": "acc.price.conf.wizard",
            "target": "new",
            "res_id": wiz.id,
            "view_id": view_id,
            "context": self.env.context,
        }

    def open_x2m_matrix(self):
        """
        Appelle la vue permettant l'affichage de la matrice
        """
        return self._open_x2m_matrix("x2many_2d_matrix_sale")

    # ------------------------------------------------------
    # Financial method APP
    # ------------------------------------------------------
    def get_financial_monitoring(self, type, prm_id=None):
        date_first = fields.Date.today() - relativedelta(months=1)
        date_end = date_utils.end_of(date_first, "month")

        if prm_id:
            acc_counter_ids = self.env["acc.counter"].browse(prm_id)
        else:
            if type == "cons":
                acc_counter_ids = self.acc_delivery_ids
            else:
                acc_counter_ids = self.acc_injection_ids

        # Consommation
        power_tab_by_cons = self.env["acc.enedis.cdc"].read_group(
            [
                ("acc_operation_id", "=", self.id),
                ("date_slot", "<=", date_end),
                ("comp_data_type", "=", "cons"),
                ("acc_counter_id", "in", acc_counter_ids.ids),
            ],
            ["power", "acc_counter_id", "date_slot"],
            ["date_slot:month", "acc_counter_id"],
            orderby="acc_counter_id",
            lazy=False,
        )

        return power_tab_by_cons
