# Copyright 2023 Le Filament (<http://www.le-filament.com>)
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).

from dateutil.relativedelta import relativedelta

from odoo import _, api, fields, models
from odoo.exceptions import ValidationError


class AccPriceConf(models.Model):
    _name = "acc.price.conf"
    _description = "Gestion des prix"

    def _get_default_currency_id(self):
        return self.env.company.currency_id.id

    # ------------------------------------------------------
    # Fields declaration
    # ------------------------------------------------------
    acc_operation_id = fields.Many2one("acc.operation", "Opération", ondelete="cascade")
    acc_injection_id = fields.Many2one(
        "acc.counter",
        domain=[("is_injection", "=", True)],
        string="Point d'injection",
        ondelete="cascade",
    )
    acc_delivery_id = fields.Many2one(
        "acc.counter",
        domain=[("is_delivery", "=", True)],
        string="Point de soutirage",
        ondelete="cascade",
    )
    acc_buyer_id = fields.Many2one(
        "res.partner",
        domain=[("is_buyer_surplus", "=", True)],
        string="Acheteur de surplus",
        ondelete="cascade",
    )
    start_date = fields.Date("Début de la période")
    end_date = fields.Date("Fin de la période")
    currency_id = fields.Many2one(
        "res.currency", "Devise", default=_get_default_currency_id
    )
    price = fields.Float("Tarif", digits="Sale Price")
    power_delivery = fields.Float(related="acc_delivery_id.power_delivery", store=True)
    zip = fields.Char(related="acc_delivery_id.zip", store=True)
    type = fields.Selection(
        [("sale", "Vente"), ("buyer", "Achat"), ("accise", "ACCISE")],
        string="Type de prix",
    )

    @api.model_create_multi
    def create(self, vals_list):
        """
        A la définition d'un nouveau tarif pour une période donnée,
        plusieurs contôles et fonctionnement:
        - Distinction entre prix de vente, prix d'achat et tarif ACCISE. Si prix
        de vente ou tarif ACCISE, création du prix pour le couple point d'injection
        point de soutirage. Si création d'un tarif achat, création du prix pour le
        couple point d'injection acheteur surplus
        - Si pas de période existante, création d'une période de début de l'existance
         du point d'injection à la date renseignée, puis création d'une période
         à partir de la date donnée, sans date de fin
        - Si période existe déjà sur les dates données, pas de création
        """
        new_vals = []
        for vals in vals_list:
            if "acc_operation_id" in vals:
                acc_operation_id = vals["acc_operation_id"]
                acc_operation = self.env["acc.operation"].browse(acc_operation_id)
            if "acc_injection_id" in vals:
                acc_injection_id = vals["acc_injection_id"]
            if "acc_delivery_id" in vals:
                acc_delivery_id = vals["acc_delivery_id"]
            else:
                acc_delivery_id = False
            if "acc_buyer_id" in vals:
                acc_buyer_id = vals["acc_buyer_id"]
            else:
                acc_buyer_id = False
            if "start_date" in vals:
                start_date = fields.Date.to_date(vals["start_date"])
            if "type" in vals:
                type = vals["type"]

            # If date start period doesn't exist
            if type == "buyer":
                price_ids = self.env["acc.price.conf"].search(
                    [
                        ("type", "=", type),
                        ("acc_operation_id", "=", acc_operation_id),
                        ("acc_buyer_id", "=", acc_buyer_id),
                        ("acc_injection_id", "=", acc_injection_id),
                    ]
                )
            else:
                price_ids = self.env["acc.price.conf"].search(
                    [
                        ("type", "=", type),
                        ("acc_operation_id", "=", acc_operation_id),
                        ("acc_delivery_id", "=", acc_delivery_id),
                        ("acc_injection_id", "=", acc_injection_id),
                    ]
                )

            if acc_operation.date_start_contract > start_date:
                raise ValidationError(
                    _("L'opération n'est pas active à cette période là")
                )

            # Creation d'une nouvelle période date
            # anniversaire à date renseignée - 1 jour
            if not price_ids and acc_operation.date_start_contract < start_date:
                val_new = {
                    "acc_operation_id": acc_operation_id,
                    "acc_injection_id": acc_injection_id,
                    "acc_delivery_id": acc_delivery_id,
                    "acc_buyer_id": acc_buyer_id,
                    "price": vals["price"],
                    "start_date": acc_operation.date_start_contract,
                    "end_date": start_date - relativedelta(days=1),
                    "type": type,
                }
                new_vals.append(val_new)

            # Si une période existe après la date renseignée
            if type == "buyer":
                price_ids = self.env["acc.price.conf"].search(
                    [
                        ("acc_operation_id", "=", acc_operation_id),
                        ("acc_buyer_id", "=", acc_buyer_id),
                        ("acc_injection_id", "=", acc_injection_id),
                        ("start_date", "<=", start_date),
                        ("end_date", ">", start_date),
                        ("type", "=", type),
                    ]
                )
            else:
                price_ids = self.env["acc.price.conf"].search(
                    [
                        ("acc_operation_id", "=", acc_operation_id),
                        ("acc_delivery_id", "=", acc_delivery_id),
                        ("acc_injection_id", "=", acc_injection_id),
                        ("start_date", "<=", start_date),
                        ("end_date", ">", start_date),
                        ("type", "=", type),
                    ]
                )
            if price_ids:
                if type == "sale":
                    message = "Un prix de vente existe déjà pour cette date là"
                elif type == "buyer":
                    message = "Un prix d'achat existe déjà pour cette date là"
                else:
                    message = "Un tarif ACCISE existe déjà pour cette date là"
                raise ValidationError(_(message))

            if type == "buyer":
                price_id = self.env["acc.price.conf"].search(
                    [
                        ("type", "=", type),
                        ("acc_operation_id", "=", acc_operation_id),
                        ("acc_buyer_id", "=", acc_buyer_id),
                        ("acc_injection_id", "=", acc_injection_id),
                    ],
                    limit=1,
                    order="create_date desc",
                )
            else:
                price_id = self.env["acc.price.conf"].search(
                    [
                        ("type", "=", type),
                        ("acc_operation_id", "=", acc_operation_id),
                        ("acc_delivery_id", "=", acc_delivery_id),
                        ("acc_injection_id", "=", acc_injection_id),
                    ],
                    limit=1,
                    order="create_date desc",
                )
            if price_id:
                price_id.end_date = start_date - relativedelta(days=1)
        vals_list += new_vals
        res = super(AccPriceConf, self).create(vals_list)
        return res
